#!/usr/bin/python3 -B
import argparse
import collections
import dataclasses
import json
import os
import pathlib
import re
import subprocess
import sys
import textwrap
from datetime import datetime
from typing import List, Dict, Iterable, Callable, Any, Tuple

from debian.debian_support import DpkgArchTable

from debputy.filesystem_scan import build_fs_from_root_dir, build_virtual_fs
from debputy.transformation_rules import TransformationRuntimeError

DEBPUTY_ROOT_DIR = pathlib.Path(__file__).parent  # TODO: Subst during install

# setup PYTHONPATH: add our installation directory.
sys.path.insert(0, str(DEBPUTY_ROOT_DIR))

from debputy._deb_options_profiles import DebOptionsAndProfiles
from debputy.architecture_support import dpkg_architecture_table
from debputy.dh_migration.migration import migrate_from_dh
from debputy.dh_migration.models import AcceptableMigrationIssues
from debputy.packages import parse_source_debian_control
from debputy.version_substitutions import load_source_variables


from debputy.deb_packaging_support import setup_control_files, build_dbgsym_package_if_relevant, \
    collect_data_from_binary_staging_dir
from debputy.highlevel_manifest import HighLevelManifest, ManifestPathInfoRuleType, ChangeDescription
from debputy.highlevel_manifest_parser import parse_manifest, ManifestParseException
from debputy.intermediate_manifest import TarMember, output_intermediate_manifest
from debputy.substitution import Substitution, SubstitutionImpl, NULL_SUBSTITUTION
from debputy.util import _error, compute_output_filename, ensure_dir, scratch_dir, resolve_source_date_epoch, _warn, \
    _normalize_path


def parse_args():
    description = textwrap.dedent(f'''\
    The `debputy` program is a manifest-based Debian packaging tool. 
    
    It is used as a part of compiling a source package and transforming it into one or
    more binary (.deb) packages.
    ''')

    parser = argparse.ArgumentParser(
        description=description,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        allow_abbrev=False,
    )
    parser.add_argument('--package-manifest', dest='package_manifest',
                        action='store', default=None,
                        help='Manifest of file metadata that is not visible in the file-system')
    parser.add_argument('--substitution', dest='substitution',
                        action='store_true', default=None,
                        help='*Some* commands can work without doing path substitution.  With this option, substitution'
                             ' is always done.  It is a no-op for commands that requires substitution'
                        )
    parser.add_argument('--no-substitution', dest='substitution',
                        action='store_false', default=None,
                        help='*Some* commands can work without doing path substitution.  With this option, substitution'
                             ' is not done.  However, this option *cannot* be used with all commands.'
                        )

    parser.add_argument('-p', '--package', dest='packages',
                        action='append', type=str, default=[],
                        help='The package(s) to act on.  Affects default permission normalization rules'
                        )

    subparsers = parser.add_subparsers(dest='command', required=True)

    generate_deb_package = subparsers.add_parser(
        'generate-debs-from-binary-staging-directories',
        help="[Internal command] Generate .deb/.udebs packages from debian/<pkg> (Not stable API)",
    )
    generate_deb_package.add_argument('output', metavar="output",
                                      help='Where to place the resulting packages.  Should be a directory'
                                      )
    # Added for "help only" - you cannot trigger this option in practice
    generate_deb_package.add_argument('--', metavar='UPSTREAM_ARGS', action='extend', nargs='+', dest='unused')

    migrate_from_debhelper = subparsers.add_parser('migrate-from-dh',
                                                   help='Generate/update manifest from a "dh $@" using package')
    migrate_from_debhelper.add_argument('--acceptable-migration-issues', dest='acceptable_migration_issues',
                                        action='append', type=str, default=[],
                                        help='Continue the migration even if this/these issues are detected.'
                                             ' Can be set to ALL (in all upper-case) to accept all issues')

    migrate_from_debhelper.add_argument('--no-act', dest='destructive',
                                        action='store_false', default=True,
                                        help='Do not perform changes.  Existing manifest will not be overridden')

    show_changes_parser = subparsers.add_parser(
        'show-manifest-rules',
        help="Show what the manifest will do (format is not stable / intended for machine parsing)",
        description=textwrap.dedent('''\
        Translate manifest into a textual description of what it would. The output is intended to be human
        readable and may change from version to version.  If you need a stable machine readable format, please
        file a feature request expanding on what data you need.
        ''')
    )

    show_changes_parser.add_argument('--show-implicit-rules', action='store_true', default=False,
                                     dest='show_implicit_rules',
                                     help="Show implicit rules (usually directories implied by other actions)")
    show_changes_parser.add_argument('--show-builtin-rules', action='store_true', default=False,
                                     dest='show_builtin_rules',
                                     help="Show builtin rules (e.g., default permission normalization)")

    generator_parser = subparsers.add_parser(
        'debug-generate-intermediate-manifest',
        help="[Internal command] Generate intermediate manifest for deb_packer (Not stable API)"
    )

    generator_parser.add_argument('package_root_dir', metavar="PACKAGE_ROOT_DIR",
                                  help='Root directory of the package. Must contain a DEBIAN directory'
                                  )
    generator_parser.add_argument('intermediate_manifest_output', nargs='?', metavar="PATH",
                                  help='Path where the intermediate manifest should be placed.'
                                  )

    trace_parser = subparsers.add_parser(
        'trace-rules',
        help="Trace what happens with a given file (format is not stable / intended for machine parsing)"
    )
    trace_parser.add_argument(
        'path',
        metavar="PATH",
        help="The path to trace. If it ends with '/' is will match as a directory, otherwise as a file."
    )

    argv = sys.argv
    try:
        i = argv.index('--')
        upstream_args = argv[i + 1:]
        argv = argv[:i]
    except (IndexError, ValueError):
        upstream_args = []
    parsed_args = parser.parse_args(argv[1:])
    setattr(parsed_args, 'upstream_args', upstream_args)
    setattr(parsed_args, 'packages', frozenset(parsed_args.packages))

    return parsed_args


def _serialize_intermediate_manifest(members: Iterable[TarMember]) -> str:
    serial_format = [m.to_manifest() for m in members]
    return json.dumps(serial_format)


def _generate_intermediate_manifest(parsed_args, manifest: HighLevelManifest, mtime: int) -> None:
    output_path = parsed_args.intermediate_manifest_output
    active_packages = list(manifest.active_packages)
    if len(active_packages) != 1:
        _error("This command requires exactly one package (-p/--package")
    root_dir = parsed_args.package_root_dir
    package = active_packages[0].name
    fs_root = build_fs_from_root_dir(root_dir)
    intermediate_manifest = manifest.apply_to_binary_staging_directory(package, fs_root, mtime)
    output_intermediate_manifest(output_path, intermediate_manifest)


def _print_manifest_change_group(grouped: Dict[ManifestPathInfoRuleType, List[ChangeDescription]],
                                 rule_type: ManifestPathInfoRuleType,
                                 header: str,
                                 footer: str = ''
                                 ):
    changes = grouped.get(rule_type)
    if changes:
        print(header)
        for change_description in changes:
            print(change_description.pretty_format())
        print()
        if footer:
            print(footer)
            print()


def _show_manifest_rules(parsed_args, manifest: HighLevelManifest, mtime: int) -> None:
    for dctrl_bin in manifest.active_packages:
        package = dctrl_bin.name
        package_state = manifest.package_state_for(package)
        declared_transformations = [t for t in package_state.transformations if not t.is_builtin_rule]

        print(f"=== {package} ===")
        print()

        if declared_transformations:
            print(f"  Transformations [{package}]:")
            for transformation in declared_transformations:
                print(f' * {transformation.describe_transformation()}')
            print()
        grouped = collections.defaultdict(list)
        for change_description in sorted(manifest.change_descriptions(package),
                                         key=lambda x: x.match_rule.sort_key(),
                                         reverse=True,
                                         ):
            path_info = change_description.path_info
            if path_info.is_builtin_definition and not parsed_args.show_builtin_rules:
                continue
            if path_info.is_implicit_definition and not parsed_args.show_implicit_rules:
                continue
            grouped[change_description.path_info.rule_type].append(change_description)

        _print_manifest_change_group(grouped, ManifestPathInfoRuleType.CREATE_DIRECTORY,
                                     f'  Directory creation [{package}]:')
        _print_manifest_change_group(grouped, ManifestPathInfoRuleType.CREATE_SYMLINK,
                                     f'  Symlink creation [{package}]:')

        print(f"  Metadata changes (default settings) [{package}]:")
        print("   * Owner: root (uid: 0)")
        print("   * Group: root (gid: 0)")
        print(f"   * mtime (clamp/upper bound): {mtime} ({datetime.utcfromtimestamp(mtime)}Z)")
        print()
        _print_manifest_change_group(grouped, ManifestPathInfoRuleType.ENSURE_METADATA,
                                     f'  Metadata changes (path rules) [{package}]:')

        if package_state.maintscript_snippets:
            print()
            print(f"  Maintscript snippets triggered by the manifest [{package}]:")
            for script, snippets in package_state.maintscript_snippets.items():
                print(f"   * {script}:")
                for snippet in snippets.all_snippets():
                    print(f"       - triggered by {snippet.definition_source}")


_RRR_DEB_ASSEMBLY_KEYWORD = 'debputy/deb-assembly'


def _determine_assembly_method(manifest: HighLevelManifest) -> Tuple[bool, List[str]]:
    if not manifest.will_assembly_with_dpkg_deb_require_root or os.getuid() == 0:
        return False, []
    rrr = os.environ.get('DEB_RULES_REQUIRES_ROOT')
    if rrr and _RRR_DEB_ASSEMBLY_KEYWORD in rrr:
        gain_root_cmd = os.environ.get('DEB_GAIN_ROOT_CMD')
        if not gain_root_cmd:
            _error('DEB_RULES_REQUIRES_ROOT contains a debputy keyword but DEB_GAIN_ROOT_CMD does not contain a '
                   '"gain root" command')
        return False, gain_root_cmd.split()
    if rrr == 'no':
        _warn('Using internal assembly method due to "Rules-Requires-Root" being "no" and dpkg-deb assembly would'
              ' require (fake)root for binary packages that needs it.')
        return True, []

    for package, def_source in manifest.package_definitions_that_require_root:
        break
    else:
        assert False
    _error(f'Due to the definition "{def_source}" in the manifest, the package assembly will require (fake)root.'
           ' However, this command is not run as root nor was debputy requested to use a root command via'
           f' "Rules-Requires-Root".  Please consider adding "{_RRR_DEB_ASSEMBLY_KEYWORD}" to "Rules-Requires-Root"'
           ' in debian/control. Though, due to #1036865, you may have to revert to'
           ' "Rules-Requires-Root: binary-targets" depending on which version of dpkg you need to support.'
           ' Alternatively, you can set "Rules-Requires-Root: no" in debian/control and debputy will assemble'
           ' the package anyway. In this case, dpkg-deb will not be used, but the output should be bit-for-bit'
           ' compatible with what debputy would have produced with dpkg-deb (and root/fakeroot).')


def _generate_debs_from_binary_staging_directories(parsed_args, manifest: HighLevelManifest, mtime: int) -> None:

    use_fallback_assembly, gain_root_cmd = _determine_assembly_method(manifest)

    output_path = parsed_args.output
    upstream_args = parsed_args.upstream_args

    package_data_table = collect_data_from_binary_staging_dir(manifest, mtime)
    scratch_root_dir = scratch_dir()
    deb_materialize = str(DEBPUTY_ROOT_DIR / 'deb_materialization.py')
    packages_needing_root = frozenset(t[0] for t in manifest.package_definitions_that_require_root)

    for dctrl_bin in manifest.active_packages:
        package = dctrl_bin.name
        dctrl_data = package_data_table[package]
        root_dir = dctrl_data.binary_staging_root_dir

        assert dctrl_bin.should_be_acted_on
        intermediate_manifest = dctrl_data.intermediate_manifest
        assert intermediate_manifest

        udeb_package_data = package_data_table.get(f'{dctrl_bin.name}-udeb')
        if udeb_package_data and not udeb_package_data.binary_package.is_udeb:
            udeb_package_data = None

        setup_control_files(dctrl_data, manifest, udeb_package_data)

        build_dbgsym_package_if_relevant(dctrl_bin, upstream_args, output_path)

        materialization_dir = os.path.join(scratch_root_dir, 'materialization-dirs', package)
        ensure_dir(os.path.dirname(materialization_dir))

        materialize_cmd = []
        if package in packages_needing_root:
            # Only use the gain_root_cmd if we absolutely need it.
            # Note that gain_root_cmd will be empty unless R³ is set to the relevant keyword
            # that would make us use targeted promotion. Therefore, we do not need to check other
            # conditions than the package needing root.
            materialize_cmd.extend(gain_root_cmd)
        materialize_cmd.extend([
            deb_materialize,
            'materialize-deb',
            '--intermediate-package-manifest', '-',
            '--source-date-epoch', str(mtime),
            '--discard-existing-output',
            root_dir,
            materialization_dir,
        ])
        output = output_path
        if dctrl_bin.is_udeb:
            materialize_cmd.append('--udeb')
            output = os.path.join(output_path, compute_output_filename(root_dir, True))

        assembly_method = 'debputy' if package in packages_needing_root and use_fallback_assembly else 'dpkg-deb'
        combined_materialization_and_assembly = package not in packages_needing_root
        if combined_materialization_and_assembly:
            materialize_cmd.extend([
                '--build-method', assembly_method,
                '--assembled-deb-output', output
            ])

        if upstream_args:
            materialize_cmd.append('--')
            materialize_cmd.extend(upstream_args)

        if combined_materialization_and_assembly:
            print(f"Materializing and assembling {package} via: {materialize_cmd}")
        else:
            print(f"Materializing {package} via: {materialize_cmd}")
        proc = subprocess.Popen(materialize_cmd, stdin=subprocess.PIPE)
        proc.communicate(_serialize_intermediate_manifest(intermediate_manifest).encode('utf-8'))
        if proc.returncode != 0:
            _error(f"{deb_materialize} exited with a non-zero exit code!")

        if not combined_materialization_and_assembly:
            build_materialization = [
                deb_materialize,
                'build-materialized-deb',
                materialization_dir,
                assembly_method,
                '--output', output,
            ]
            print(f"Assembling {package} via: {build_materialization}")
            subprocess.check_call(build_materialization)


def _migrate_from_dh(parsed_args, manifest: HighLevelManifest, _mtime: int) -> None:
    acceptable_migration_issues = AcceptableMigrationIssues(
        frozenset(i for x in parsed_args.acceptable_migration_issues for i in x.split(','))
    )
    migrate_from_dh(manifest, acceptable_migration_issues, parsed_args.destructive)


def _trace_rules(parsed_args, manifest: HighLevelManifest, _mtime: int) -> None:
    path: str = parsed_args.path
    package: str = ""
    if path.startswith(('debian/', './debian/')):
        m = re.search(r'(?:[.]/)?debian/([^/]+)/(.+)', path)
        if m:
            pkg, subpath = m.groups()
            if pkg in (p.name for p in manifest.all_packages):
                print(f"N: Assuming you meant {subpath} in package {pkg}")
                print()
                path = subpath
                package = pkg
        if not package:
            print(f"W: In the current version, the path should be relative to the binary package root.")
            print(f'W: That is, use "usr/bin" rather than "debian/foo/usr/bin".')
            print()
    if not package:
        active_packages = list(manifest.active_packages)
        if len(active_packages) != 1:
            _error("This action need exactly one package. Please use -p<package> to pick one")
        package = active_packages[0].name
    normalized_path = _normalize_path(path)
    # Restore trailing slash if _normalize_path removed it.
    normalized_path_as_path_def = normalized_path + ('/' if path.endswith('/') else '')
    fs_root = build_virtual_fs([normalized_path_as_path_def])
    fs_path = fs_root.lookup(normalized_path)
    trace_paths = {normalized_path}
    print(f"Tracing rules for: {fs_path.path} ({'directory' if fs_path.is_dir else 'file'})")
    print()

    for event in manifest.trace_apply_to_binary_staging_directory(package, fs_root):
        if event.fs_path.path not in trace_paths:
            continue
        if not event.is_active:
            conditional_description = event.condition.describe() if event.condition else '<Unknown conditional>'
            print(f"   - One match ignored due to a unsatisfied condition: {conditional_description}")
        else:
            print(f"{event.describe_event()}")

    print()
    print(f"W: Transformation rules (packages.{package}.transformations in the manifest) are currently not traced.")
    print(f"W: Any entry that list (if it exists) is currently silently ignored.")


@dataclasses.dataclass
class Command:

    handler: Callable[[Any, HighLevelManifest, int], None]
    require_substitution: bool = True

    def create_substitution(self, parsed_args, extra_substitutions: Dict[str, str]) -> Substitution:
        requested_subst = parsed_args.substitution
        if requested_subst is False and self.require_substitution:
            _error(f"--no-substitution cannot be used with {parsed_args.command}")
        if self.require_substitution or requested_subst is not False:
            if not os.path.isfile('debian/changelog'):
                _error('Expected to be run from a directory containing debian/changelog')
            return SubstitutionImpl(extra_substitutions=extra_substitutions)
        return NULL_SUBSTITUTION


COMMANDS = {
    'debug-generate-intermediate-manifest': Command(handler=_generate_intermediate_manifest),
    'generate-debs-from-binary-staging-directories': Command(handler=_generate_debs_from_binary_staging_directories),
    'show-manifest-rules': Command(handler=_show_manifest_rules, require_substitution=False),
    'migrate-from-dh': Command(handler=_migrate_from_dh),
    'trace-rules': Command(handler=_trace_rules),
}


def main():
    parsed_args = parse_args()

    command = COMMANDS[parsed_args.command]

    packages = parsed_args.packages
    v = load_source_variables('debian/changelog')
    substitution = command.create_substitution(parsed_args, v)
    mtime = resolve_source_date_epoch(None)
    os.environ['SOURCE_DATE_EPOCH'] = v['SOURCE_DATE_EPOCH']
    build_env = DebOptionsAndProfiles.instance()
    dpkg_architecture_variables = dpkg_architecture_table()
    dpkg_arch_query_table = DpkgArchTable.load_arch_table()

    _, binary_packages = parse_source_debian_control(
        packages,  # -p/--package
        set(),  # -N/--no-package
        False,  # -i
        False,  # -a
        dpkg_architecture_variables=dpkg_architecture_variables,
        dpkg_arch_query_table=dpkg_arch_query_table,
        build_env=build_env,
    )
    assert packages <= binary_packages.keys()

    try:
        manifest = parse_manifest(parsed_args.package_manifest,
                                  binary_packages,
                                  substitution,
                                  dpkg_architecture_variables,
                                  dpkg_arch_query_table,
                                  build_env,
                                  )
    except (ManifestParseException, TransformationRuntimeError) as e:
        _error(e.args[0])
    command.handler(parsed_args, manifest, mtime)


if __name__ == '__main__':
    main()
