import dataclasses
import subprocess
from typing import Iterable, Optional, List, Dict, Mapping, Tuple

from debputy.filesystem_scan import FSPath
from debputy.intermediate_manifest import PathType
from debputy.manifest_path_rules import ManifestPathRule
from debputy.packager_provided_files import PACKAGER_PROVIDED_FILE_DEFINITIONS, PackagerProvidedFileClass
from debputy.packages import BinaryPackage
from debputy.path_matcher import MATCH_ANYTHING, MatchRule, ExactFileSystemPath, DirectoryBasedMatch, MatchRuleType, \
    BasenameGlobMatch
from debputy.substitution import Substitution
from debputy.typing import DirEntryLike
from debputy.util import _normalize_path

# Imported from dh_fixperms
_PERMISSION_NORMALIZATION_SOURCE_DEFINITION = 'permission normalization'
_STD_FILE_MODE = 0o644
_PATH_FILE_MODE = 0o755
_PERL_MODULE_DIRS: Optional[List[str]] = None

DEFAULT_PATH_INFO = ManifestPathRule.ensure_metadata(
    _PERMISSION_NORMALIZATION_SOURCE_DEFINITION,
    MATCH_ANYTHING,
    uname='root',
    uid=0,
    gname='root',
    gid=0,
    # This mode is from dh_fixperms
    symbolic_mode='go=rX,u+rw,a-s',
    is_builtin_definition=True,
)


def _perl_module_dirs() -> Iterable[str]:
    global _PERL_MODULE_DIRS
    if _PERL_MODULE_DIRS is None:
        output = subprocess.check_output(
            ['perl', '-MConfig', '-e', 'print "$Config{vendorlib}\n$Config{vendorarch}\n"']
        )
        _PERL_MODULE_DIRS = [
            _normalize_path(path)
            for path in output.decode('utf-8').splitlines(keepends=False)
        ]
    yield from _PERL_MODULE_DIRS


class _UsrShareDocMatchRule(DirectoryBasedMatch):

    def __init__(self) -> None:
        super().__init__(MatchRuleType.ANYTHING_BENEATH_DIR,
                         _normalize_path('usr/share/doc', with_prefix=True),
                         path_type=PathType.FILE,
                         )

    def finditer(self, fs_root: 'FSPath') -> Iterable['FSPath']:
        doc_dir = fs_root.lookup(self._directory)
        if doc_dir is None:
            return
        for path_in_doc_dir in doc_dir.children:
            yield path_in_doc_dir
            for subpath in path_in_doc_dir.children:
                if subpath.basename == 'examples' and subpath.is_dir:
                    continue
                yield from subpath.all_paths()

    def describe_match_short(self) -> str:
        return f'All files beneath {self._directory}/ except .../<pkg>/examples'

    def describe_match_exact(self) -> str:
        return self.describe_match_short()

    def __hash__(self):
        return object.__hash__(self)

    def __eq__(self, other: object) -> bool:
        return self is other


USR_SHARE_DOC_MATCH_RULE = _UsrShareDocMatchRule()
del _UsrShareDocMatchRule


@dataclasses.dataclass(frozen=True, slots=True)
class PackagerProvidedFile:
    source_basename: str
    package_name: str
    installed_as_basename: str
    provided_key: str
    definition: PackagerProvidedFileClass
    match_priority: int = 0

    def compute_dest_name(self) -> str:
        return self.definition.compute_dest_name(self.installed_as_basename)


def _find_package_name_prefix(binary_packages: Mapping[str, BinaryPackage],
                              main_binary_package: str,
                              max_periods_in_package_name: int,
                              orig_basename: str,
                              ) -> Iterable[Tuple[str, str, bool]]:
    if max_periods_in_package_name < 1:
        prefix, remaining = orig_basename.split('.', 1)
        if prefix in binary_packages:
            yield prefix, remaining, True
        else:
            yield main_binary_package, orig_basename, False
        return

    parts = orig_basename.split('.', max_periods_in_package_name + 1)
    for p in range(len(parts) - 1, 0, -1):
        name = '.'.join(parts[0:p])
        if name in binary_packages:
            remaining = '.'.join(parts[p:])
            yield name, remaining, True
    # main package case
    yield main_binary_package, orig_basename, False


def _split_path(binary_packages: Mapping[str, BinaryPackage],
                main_binary_package: str,
                max_periods_in_package_name: int,
                orig_basename: str,
                ) -> Optional[PackagerProvidedFile]:
    owning_package_name = main_binary_package
    basename = orig_basename
    match_priority = 0
    if '.' not in basename:
        definition = PACKAGER_PROVIDED_FILE_DEFINITIONS.get(basename)
        if definition is None:
            return None
        return PackagerProvidedFile(
            source_basename=orig_basename,
            package_name=owning_package_name,
            installed_as_basename=owning_package_name,
            provided_key='.UNNAMED.',
            definition=definition,
            match_priority=match_priority,
        )

    for owning_package_name, basename, explicit_package in _find_package_name_prefix(binary_packages,
                                                                                     main_binary_package,
                                                                                     max_periods_in_package_name,
                                                                                     orig_basename,
                                                                                     ):
        owning_package = binary_packages[owning_package_name]
        match_priority = 1 if explicit_package else 0
        if not owning_package.is_arch_all and '.' in basename:
            remaining, last_word = basename.rsplit('.', 1)
            if last_word == owning_package.resolved_architecture:
                match_priority = 3
                basename = remaining
            elif last_word == owning_package.package_deb_architecture_variable('ARCH_OS'):
                match_priority = 2
                basename = remaining

        definition = PACKAGER_PROVIDED_FILE_DEFINITIONS.get(basename)
        if definition is not None:
            return PackagerProvidedFile(
                source_basename=orig_basename,
                package_name=owning_package_name,
                installed_as_basename=owning_package_name,
                provided_key='.UNNAMED.',
                definition=definition,
                match_priority=match_priority,
            )

        install_as_name = basename
        file_class = ''
        while '.' in install_as_name:
            install_as_name, file_class_part = install_as_name.rsplit('.', 1)
            file_class = file_class_part + '.' + file_class if file_class != '' else file_class_part
            definition = PACKAGER_PROVIDED_FILE_DEFINITIONS.get(file_class)
            if definition is not None:
                return PackagerProvidedFile(
                    source_basename=orig_basename,
                    package_name=owning_package_name,
                    installed_as_basename=install_as_name,
                    provided_key=install_as_name,
                    definition=definition,
                    match_priority=match_priority,
                )
    return None


def builtin_packager_provided_files(debian_dir: Iterable[DirEntryLike],
                                    binary_packages: Mapping[str, BinaryPackage],
                                    ) -> Dict[str, List['ManifestPathRule']]:
    main_binary_package = [p.name for p in binary_packages.values() if p.is_main_package][0]
    provided_files: Dict[str, Dict[str, PackagerProvidedFile]] = {n: {} for n in binary_packages}
    max_periods_in_package_name = max(name.count('.') for name in binary_packages)

    for entry in debian_dir:
        if not entry.is_file(follow_symlinks=True):
            continue
        packager_provided_file = _split_path(binary_packages,
                                             main_binary_package,
                                             max_periods_in_package_name,
                                             entry.name)
        if packager_provided_file is None:
            continue
        provided_files_for_package = provided_files[packager_provided_file.package_name]
        existing = provided_files_for_package.get(packager_provided_file.provided_key)
        if existing is not None and existing.match_priority > packager_provided_file.match_priority:
            continue
        provided_files_for_package[packager_provided_file.provided_key] = packager_provided_file

    result = {
        k: [
            ManifestPathRule.packager_provided_file(
                definition_source=f'debian/{x.source_basename} (auto-detected by debputy)',
                match_rule=ExactFileSystemPath(
                    f'{x.definition.install_into_directory}/{x.compute_dest_name()}'
                ),
                source_path=f'debian/{x.source_basename}',
                mode=x.definition.default_mode,
                is_builtin_definition=True,
            ) for x in v.values()
        ] for k, v in provided_files.items()
    }

    return result


def builtin_path_metadata_rules(substitution: Substitution) -> Iterable['ManifestPathRule']:
    yield from (
        ManifestPathRule.ensure_metadata(
            _PERMISSION_NORMALIZATION_SOURCE_DEFINITION,
            MatchRule.from_path_or_glob(x, _PERMISSION_NORMALIZATION_SOURCE_DEFINITION, path_type=PathType.FILE),
            mode=_STD_FILE_MODE,
            is_builtin_definition=True,
        )
        for x in ('*.so.*', '*.so', '*.la', '*.a', '*.js', '*.css', '*.scss', '*.sass', '*.jpeg', '*.jpg',
                  '*.png', '*.gif', '*.cmxs', '*.node',)
    )

    yield from (
        ManifestPathRule.ensure_metadata(
            _PERMISSION_NORMALIZATION_SOURCE_DEFINITION,
            MatchRule.recursive_beneath_directory(x,
                                                  _PERMISSION_NORMALIZATION_SOURCE_DEFINITION,
                                                  path_type=PathType.FILE),
            mode=_STD_FILE_MODE,
            is_builtin_definition=True,
        )
        for x in ('usr/share/man', 'usr/include', 'usr/share/applications', 'usr/share/lintian/overrides',)
    )

    # The dh_fixperms tool recuses for these directories, but probably should not (see #1006927)
    yield from (
        ManifestPathRule.ensure_metadata(
            _PERMISSION_NORMALIZATION_SOURCE_DEFINITION,
            MatchRule.from_path_or_glob(f"{x}/*",
                                        _PERMISSION_NORMALIZATION_SOURCE_DEFINITION,
                                        path_type=PathType.FILE),
            mode=_PATH_FILE_MODE,
            is_builtin_definition=True,
        )
        for x in ('usr/bin', 'usr/bin/mh', 'bin', 'usr/sbin', 'sbin', 'usr/games', 'usr/libexec', 'etc/init.d',)
    )
    yield from (
        # Strictly speaking, dh_fixperms does a recursive search but in practice, it does not matter.
        ManifestPathRule.ensure_metadata(
            _PERMISSION_NORMALIZATION_SOURCE_DEFINITION,
            MatchRule.from_path_or_glob(x, _PERMISSION_NORMALIZATION_SOURCE_DEFINITION, path_type=PathType.FILE),
            mode=0o440,
            is_builtin_definition=True,
        )
        for x in ('etc/sudoers.d/*',)
    )

    # The reportbug rule
    yield ManifestPathRule.ensure_metadata(
        _PERMISSION_NORMALIZATION_SOURCE_DEFINITION,
        ExactFileSystemPath(substitution.substitute(_normalize_path("usr/share/bug/{{PACKAGE}}"),
                                                    _PERMISSION_NORMALIZATION_SOURCE_DEFINITION)),
        mode=0o755,
        dirmode=0o755,
        is_builtin_definition=True,
    )
    yield ManifestPathRule.ensure_metadata(
        _PERMISSION_NORMALIZATION_SOURCE_DEFINITION,
        MatchRule.recursive_beneath_directory("usr/share/bug/{{PACKAGE}}",
                                              _PERMISSION_NORMALIZATION_SOURCE_DEFINITION,
                                              path_type=PathType.FILE,
                                              substitution=substitution,
                                              ),
        mode=0o644,
        is_builtin_definition=True,
    )
    yield ManifestPathRule.ensure_metadata(
        _PERMISSION_NORMALIZATION_SOURCE_DEFINITION,
        ExactFileSystemPath(substitution.substitute(_normalize_path("usr/share/bug/{{PACKAGE}}/script"),
                                                    _PERMISSION_NORMALIZATION_SOURCE_DEFINITION)),
        mode=0o755,
        dirmode=0o755,
        is_builtin_definition=True,
    )
    yield ManifestPathRule.ensure_metadata(
        _PERMISSION_NORMALIZATION_SOURCE_DEFINITION,
        USR_SHARE_DOC_MATCH_RULE,
        mode=0o0644,
        is_builtin_definition=True,
    )
    yield from (
        ManifestPathRule.ensure_metadata(
            _PERMISSION_NORMALIZATION_SOURCE_DEFINITION,
            BasenameGlobMatch('*.pm',
                              only_when_in_directory=perl_dir,
                              path_type=PathType.FILE,
                              recursive_match=True,
                              ),
            symbolic_mode='a-x',
            is_builtin_definition=True,
        ) for perl_dir in _perl_module_dirs()
    )
    yield ManifestPathRule.ensure_metadata(
        _PERMISSION_NORMALIZATION_SOURCE_DEFINITION,
        BasenameGlobMatch('*.ali',
                          only_when_in_directory=_normalize_path('usr/lib'),
                          path_type=PathType.FILE,
                          recursive_match=True,
                          ),
        symbolic_mode='a-w',
        is_builtin_definition=True,
    )
