import collections
import contextlib
import dataclasses
import hashlib
import itertools
import os
import subprocess
import sys
import textwrap
from typing import Iterable, List, Optional, Set, Any, Generator, Dict, Sequence

from debian.substvars import Substvars

from debputy.debhelper_emulation import dhe_install_pkg_file_as_ctrl_file_if_present, dhe_dbgsym_root_dir, \
    dhe_read_dbgsym_build_ids, dhe_read_dbgsym_migration
from debputy.filesystem_scan import FSPath, build_fs_from_root_dir
from debputy.highlevel_manifest import HighLevelManifest, PackageTransformationDefinition
from debputy.installdeb_emulation import dh_installdeb_emulation
from debputy.intermediate_manifest import IntermediateManifest
from debputy.maintscript_snippet import UDEB_CONTROL_SCRIPTS, MaintscriptSnippet, ALL_CONTROL_SCRIPTS
from debputy.packager_provided_files import PackagerProvidedFileClass
from debputy.packages import BinaryPackage
from debputy.packaging.makeshlibs import compute_shlibs, SONAMEInfo, extract_soname_info
from debputy.packaging.shlibdeps import find_shlibdeps
from debputy.util import _error, ensure_dir, escape_shell, POSTINST_DEFAULT_CONDITION, assume_not_none

# For unclean builds, we need to unlink these to generate reproducible output.
_UNCLEAN_BUILD_UNLINK_CONTROL_FILES = [
    'control',
    'md5sums',
    'conffiles',
]


@dataclasses.dataclass(slots=True, frozen=True)
class BinaryPackageData:

    binary_package: BinaryPackage
    binary_staging_root_dir: str
    fs_root: Optional[FSPath]
    intermediate_manifest: Optional[IntermediateManifest]
    shared_libraries: Optional[List[SONAMEInfo]]


def generate_md5sums_file(root_dir: str, fs_root: FSPath) -> None:
    conffiles = os.path.join(root_dir, 'DEBIAN', 'conffiles')
    md5sums = os.path.join(root_dir, 'DEBIAN', 'md5sums')
    exclude = set()
    if os.path.isfile(conffiles):
        with open(conffiles, 'rt') as fd:
            for line in fd:
                if not line.startswith("/"):
                    continue
                exclude.add('.' + line.rstrip("\n"))
    had_content = False
    files = (path for path in fs_root.all_paths() if path.is_file and path.path not in exclude)
    with open(md5sums, 'wt') as md5fd:
        for member in files:
            path = member.path
            assert path.startswith('./')
            path = path[2:]
            with open(member.fs_path, 'rb') as f:
                file_hash = hashlib.md5()
                while chunk := f.read(8192):
                    file_hash.update(chunk)
            had_content = True
            md5fd.write(f"{file_hash.hexdigest()}  {path}\n")
    if not had_content:
        os.unlink(md5sums)


def install_or_generate_conffiles(binary_package: BinaryPackage, root_dir: str, fs_root: FSPath) -> None:
    conffiles_dest = os.path.join(root_dir, 'DEBIAN', 'conffiles')
    dhe_install_pkg_file_as_ctrl_file_if_present(binary_package, 'conffiles', root_dir, 0o0644)
    etc_dir = fs_root.lookup('etc')
    if etc_dir:
        _add_conffiles(conffiles_dest, (p for p in etc_dir.all_paths() if p.is_file))
    if os.path.isfile(conffiles_dest):
        os.chmod(conffiles_dest, 0o0644)


def collect_data_from_binary_staging_dir(manifest: HighLevelManifest, mtime: int) -> Dict[str, BinaryPackageData]:
    package_data_table = {}

    for dctrl_bin in manifest.all_packages:
        package = dctrl_bin.name
        root_dir = os.path.join('debian', package)
        fs_root = None
        intermediate_manifest = None
        sonames = None
        if dctrl_bin.should_be_acted_on:
            fs_root = build_fs_from_root_dir(root_dir)
            intermediate_manifest = manifest.apply_to_binary_staging_directory(package, fs_root, mtime)
            sonames = extract_soname_info(dctrl_bin, fs_root)

        package_data_table[package] = BinaryPackageData(
            dctrl_bin,
            root_dir,
            fs_root,
            intermediate_manifest,
            sonames,
        )

    return package_data_table


def build_dbgsym_package_if_relevant(binary_package: BinaryPackage,
                                     dpkg_deb_args: List[str],
                                     output_dir: str,
                                     ) -> None:
    if binary_package.is_udeb:
        # We never built udebs due to #797391.  If it is easy to do, we can fix it when we do the control file
        return
    if not os.path.isdir(output_dir):
        _error(f"Cannot produce a dbgsym package when output path is not a directory.")
        return
    dbgsym_root = dhe_dbgsym_root_dir(binary_package)
    if not os.path.isdir(dbgsym_root):
        return
    package = binary_package.name
    if not os.path.isfile(os.path.join(dbgsym_root, 'DEBIAN', 'control')):
        # Will not be a problem when we integrate generation of DEBIAN/control (#18)
        _error(f"There is a dbgsym directory from debhelper for {package}, but not a DEBIAN/control")

    fs_root = build_fs_from_root_dir(dbgsym_root)
    generate_md5sums_file(dbgsym_root, fs_root)

    dpkg_cmd = [
        'dpkg-deb',
        '--root-owner-group',
    ]
    dpkg_cmd.extend(dpkg_deb_args)
    dpkg_cmd.extend([
        '--build',
        dbgsym_root,
        output_dir,
    ])

    print(f"Executing for {package}-dbgsym: {dpkg_cmd}")
    subprocess.check_call(dpkg_cmd)


def _register_directories_with_update_icons_cache(
        fs_root: FSPath,
        maintscript_snippets: Dict[str, List[MaintscriptSnippet]],
        ) -> None:
    icons_root_dir = fs_root.lookup('./usr/share/icons')
    if not icons_root_dir:
        return
    icon_dirs = []
    for subdir in icons_root_dir.children:
        if subdir.basename in ('gnome', 'hicolor'):
            # dh_icons skips this for some reason.
            continue
        for p in subdir.all_paths():
            if p.is_file and p.basename.endswith(('.png', '.svg', '.xpm', '.icon')):
                icon_dirs.append(subdir.path.lstrip('.'))
                break
    if not icon_dirs:
        return

    icon_dir_list_escaped = escape_shell(*icon_dirs)

    postinst_snippet = textwrap.dedent(f'''\
        if {POSTINST_DEFAULT_CONDITION}; then
            if command -v update-icon-caches >/dev/null; then
                update-icon-caches {icon_dir_list_escaped}
            fi
        fi
    ''')

    postrm_snippet = textwrap.dedent(f'''\
        if command -v update-icon-caches >/dev/null; then
            update-icon-caches {icon_dir_list_escaped}
        fi
    ''')

    maintscript_snippets['postinst'].append(MaintscriptSnippet(
        'Auto-detected by debputy',
        postinst_snippet,
    ))

    maintscript_snippets['postrm'].append(MaintscriptSnippet(
        'Auto-detected by debputy',
        postrm_snippet,
    ))


def _register_gsettings_dependencies(fs_root: FSPath,
                                     substvar_dependencies: Dict[str, Set[str]],
                                     ) -> None:

    gsettings_schema_dir = fs_root.lookup(PackagerProvidedFileClass.GSETTINGS_OVERRIDE_FILE.install_into_directory)
    if not gsettings_schema_dir:
        return

    for path in gsettings_schema_dir.all_paths():
        if path.is_file and path.basename.endswith(('.xml', '.override')):
            substvar_dependencies['misc:Depends'].add('dconf-gsettings-backend | gsettings-backend')
            break


def _register_xfonts_snippets(
        package_name: str,
        fs_root: FSPath,
        maintscript_snippets: Dict[str, List[MaintscriptSnippet]],
        substvar_dependencies: Dict[str, Set[str]],
        ) -> None:

    xfonts_root_dir = fs_root.lookup('./usr/share/fonts/X11/')
    if not xfonts_root_dir:
        return

    cmds = []
    cmds_postinst = []
    cmds_postrm = []

    for xfonts_dir in xfonts_root_dir.children:
        xfonts_dirname = xfonts_dir.basename
        if not xfonts_dir.is_dir and not xfonts_dirname.startswith('.'):
            continue
        if fs_root.lookup(f'./etc/X11/xfonts/{xfonts_dirname}/{package_name}.scale'):
            cmds.append(escape_shell('update-fonts-scale', xfonts_dirname))
        cmds.append(escape_shell('update-fonts-dir', '--x11r7-layout', xfonts_dirname))
        alias_file = f'etc/X11/xfonts/{xfonts_dirname}/{package_name}.alias'
        if fs_root.lookup(f'./{alias_file}'):
            cmds_postinst.append(escape_shell('update-fonts-alias', '--include', f'/{alias_file}', xfonts_dirname))
            cmds_postrm.append(escape_shell('update-fonts-alias', '--exclude', f'/{alias_file}', xfonts_dirname))

    if not cmds:
        return

    postinst_snippet = textwrap.dedent(f'''\
        if command -v update-fonts-dir >/dev/null; then
            {';'.join(itertools.chain(cmds, cmds_postinst))}
        fi
    ''')

    postrm_snippet = textwrap.dedent(f'''\
        if [ -x "`command -v update-fonts-dir`" ]; then
            {';'.join(itertools.chain(cmds, cmds_postrm))}
        fi
    ''')

    maintscript_snippets['postinst'].append(MaintscriptSnippet(
        'Auto-detected by debputy',
        postinst_snippet,
    ))

    maintscript_snippets['postrm'].append(MaintscriptSnippet(
        'Auto-detected by debputy',
        postrm_snippet,
    ))

    substvar_dependencies['misc:Depends'].add('xfonts-utils')


def _all_tmpfiles_conf(fs_root: FSPath) -> Iterable[FSPath]:
    seen_tmpfiles = set()
    tmpfiles_dirs = [
        PackagerProvidedFileClass.SYSTEMD_TMPFILES_FILE.install_into_directory,
        './etc/tmpfiles.d',
    ]
    for tmpfiles_dir_path in tmpfiles_dirs:
        tmpfiles_dir = fs_root.lookup(tmpfiles_dir_path)
        if not tmpfiles_dir:
            continue
        for path in tmpfiles_dir.children:
            if not path.is_file or not path.basename.endswith('.conf') or path.basename in seen_tmpfiles:
                continue
            seen_tmpfiles.add(path.basename)
            yield path


def _register_systemd_tmpfiles(
        fs_root: FSPath,
        maintscript_snippets: Dict[str, List[MaintscriptSnippet]],
        ) -> None:
    tmpfiles_confs = [x.path.lstrip('.') for x in sorted(_all_tmpfiles_conf(fs_root), key=lambda x: x.basename)]
    if not tmpfiles_confs:
        return

    tmpfiles_escaped = escape_shell(*tmpfiles_confs)

    postinst_snippet = textwrap.dedent(f'''\
        if {POSTINST_DEFAULT_CONDITION}; then
            if [ -x "$(command -v systemd-tmpfiles)" ]; then
                systemd-tmpfiles {{{{DPKG_ROOT:+--root="$DPKG_ROOT"}}}} --create {tmpfiles_escaped} >/dev/null || true
            fi
        fi
    ''')

    maintscript_snippets['postinst'].append(MaintscriptSnippet(
        'Auto-detected by debputy',
        postinst_snippet,
    ))


def _register_initramfs_hooks(fs_root: FSPath, generated_triggers: Set[str]) -> None:
    hook_dir_path = PackagerProvidedFileClass.INITRAMFS_HOOK_FILE.install_into_directory
    hook_dir = fs_root.lookup(hook_dir_path)
    if not hook_dir:
        return
    for _ in hook_dir.children:
        # Only add the trigger if the directory is non-empty. It is unlikely to matter a lot,
        # but we do this to match debhelper.
        break
    else:
        return

    generated_triggers.add('activate-noawait update-initramfs')


KERNEL_MODULE_EXTENSIONS = tuple(
    f'{ext}{comp_ext}'
    for ext, comp_ext in itertools.product(
        ('.o', '.ko'),
        ('', '.gz', '.bz2', '.xz'),
    )
)


def _register_kernel_modules(fs_root: FSPath,
                             maintscript_snippets: Dict[str, List[MaintscriptSnippet]],
                             ) -> None:
    module_root_dir = fs_root.lookup('./lib/modules')

    if not module_root_dir:
        return

    module_version_dirs = []

    for module_version_dir in module_root_dir.children:
        if not module_version_dir.is_dir:
            continue

        for fs_path in module_version_dir.all_paths():
            if fs_path.basename.endswith(KERNEL_MODULE_EXTENSIONS):
                module_version_dirs.append(module_version_dir.basename)
                break

    for module_version in module_version_dirs:
        module_version_escaped = escape_shell(module_version)
        postinst_snippet = textwrap.dedent(f'''\
            if {POSTINST_DEFAULT_CONDITION}; then
                if [ -e /boot/System.map-{module_version_escaped} ]; then
                    depmod -a -F /boot/System.map-{module_version_escaped} {module_version_escaped} || true
                fi
            fi
        ''')

        postrm_snippet = textwrap.dedent(f'''\
            if [ -e /boot/System.map-{module_version_escaped} ]; then
                depmod -a -F /boot/System.map-{module_version_escaped} {module_version_escaped} || true
            fi
        ''')

        maintscript_snippets['postinst'].append(MaintscriptSnippet(
            'Auto-detected by debputy',
            postinst_snippet,
        ))

        maintscript_snippets['postrm'].append(MaintscriptSnippet(
            'Auto-detected by debputy',
            postrm_snippet,
        ))


def setup_control_files(binary_package_data: BinaryPackageData,
                        manifest: HighLevelManifest,
                        related_udeb_package_data: Optional[BinaryPackageData],
                        ) -> None:
    binary_package = assume_not_none(binary_package_data.binary_package)
    root_dir = binary_package_data.binary_staging_root_dir
    fs_root = assume_not_none(binary_package_data.fs_root)
    soname_info_list = assume_not_none(binary_package_data.shared_libraries)
    ensure_dir(os.path.join(root_dir, 'DEBIAN'))
    package_state = manifest.package_state_for(binary_package.name)
    substvar_dependencies = collections.defaultdict(set)
    substvars_file = f'debian/{binary_package.name}.substvars'

    dbgsym_root_dir = dhe_dbgsym_root_dir(binary_package)

    for ctrl_file_to_be_removed in _UNCLEAN_BUILD_UNLINK_CONTROL_FILES:
        try:
            os.unlink(os.path.join(root_dir, 'DEBIAN', ctrl_file_to_be_removed))
        except FileNotFoundError:
            pass
        try:
            os.unlink(os.path.join(dbgsym_root_dir, 'DEBIAN', ctrl_file_to_be_removed))
        except FileNotFoundError:
            pass

    generated_triggers = set()

    if not binary_package.is_arch_all:
        udeb_package_name = related_udeb_package_data.binary_package.name if related_udeb_package_data else None
        compute_shlibs(binary_package,
                       root_dir,
                       fs_root,
                       manifest,
                       soname_info_list,
                       udeb_package_name,
                       generated_triggers,
                       )
        find_shlibdeps(binary_package, fs_root, substvars_file)

    if binary_package.is_udeb:
        _generate_control_files(binary_package, package_state, root_dir, fs_root, substvars_file, substvar_dependencies)
        for script in UDEB_CONTROL_SCRIPTS:
            dhe_install_pkg_file_as_ctrl_file_if_present(binary_package, script, root_dir, 0o0755)
        return

    state = manifest.package_state_for(binary_package.name)

    maintscript_snippets = {k: v.copy() for k, v in state.maintscript_snippets.items()}
    _register_systemd_tmpfiles(fs_root, maintscript_snippets)
    _register_kernel_modules(fs_root, maintscript_snippets)
    _register_directories_with_update_icons_cache(fs_root, maintscript_snippets)
    _register_gsettings_dependencies(fs_root, substvar_dependencies)
    _register_xfonts_snippets(binary_package.name, fs_root, maintscript_snippets, substvar_dependencies)
    _register_initramfs_hooks(fs_root, generated_triggers)

    dh_installdeb_emulation(binary_package, root_dir, generated_triggers, maintscript_snippets)
    install_or_generate_conffiles(binary_package, root_dir, fs_root)
    _generate_control_files(binary_package, package_state, root_dir, fs_root, substvars_file, substvar_dependencies)
    generate_md5sums_file(root_dir, fs_root)


def _add_conffiles(conffiles_dest: str, conffile_matches: Iterable[FSPath]) -> None:
    with open(conffiles_dest, 'at') as fd:
        for conffile_match in conffile_matches:
            conffile = conffile_match.path
            assert conffile_match.is_file
            assert conffile.startswith('./')
            fd.write(f"{conffile[1:]}\n")
    if os.stat(conffiles_dest).st_size == 0:
        os.utime(conffiles_dest)


@contextlib.contextmanager
def _ensure_base_substvars_defined(substvars_file: str) -> Generator[Substvars, Any, Any]:
    with Substvars.load_from_path(substvars_file, missing_ok=True) as substvars:
        for substvar in ('misc:Depends', 'misc:Pre-Depends'):
            if substvar not in substvars:
                substvars[substvar] = ''
        yield substvars


def _compute_installed_size(fs_root: FSPath) -> int:
    """Emulate dpkg-gencontrol's code for computing the default Installed-Size"""
    size_in_kb = 0
    hard_links = set()
    for path in fs_root.all_paths():
        if not path.is_dir and path.has_fs_path:
            st = path.stat()
            if st.st_nlink > 1:
                hl_key = (st.st_dev, st.st_ino)
                if hl_key in hard_links:
                    continue
                hard_links.add(hl_key)
            path_size = (st.st_size + 1023) // 1024
        elif path.is_symlink:
            path_size = (len(path.readlink()) + 1023) // 1024
        else:
            path_size = 1
        size_in_kb += path_size
    return size_in_kb


def _generate_dbgsym_control_file_if_relevant(binary_package: BinaryPackage,
                                              dbgsym_root_dir: str,
                                              substvars_file: str,
                                              dbgsym_ids: str,
                                              multi_arch: Optional[str],
                                              extra_common_params: Sequence[str],
                                              ) -> None:
    section = binary_package.archive_section
    component = ''
    extra_params = []
    migration = dhe_read_dbgsym_migration(binary_package)
    if section is not None and '/' in section and not section.startswith('main/'):
        component = section.split('/', 1)[1] + '/'
    if multi_arch != 'same':
        extra_params.append('-UMulti-Arch')
    if migration:
        extra_params.append(f"-DReplaces={migration}")
        extra_params.append(f"-DBreaks={migration}")
    else:
        extra_params.append("-UReplaces")
        extra_params.append("-UBreaks")
    extra_params.extend(extra_common_params)

    ensure_dir(os.path.join(dbgsym_root_dir, 'DEBIAN'))
    package = binary_package.name
    dpkg_cmd = [
        'dpkg-gencontrol',
        f'-p{package}',
        '-ldebian/changelog',
        f'-T{substvars_file}',
        f'-P{dbgsym_root_dir}',
        f'-DPackage={package}-dbgsym',
        '-DDepends=' + package + ' (= ${binary:Version})',
        f'-DDescription=debug symbols for {package}',
        f'-DSection={component}debug',
        f'-DBuild-Ids={dbgsym_ids}',
        '-UPre-Depends',
        '-URecommends',
        '-USuggests',
        '-UEnhances',
        '-UProvides',
        '-UEssential',
        '-UConflicts',
        '-DPriority=optional',
        '-UHomepage',
        '-UImportant',
        '-UBuilt-Using',
        '-DAuto-Built-Package=debug-symbols',
        '-UProtected',
        *extra_params,

    ]
    print(f'Executing for {package}-dbgsym: {dpkg_cmd}')
    subprocess.check_call(dpkg_cmd)
    os.chmod(os.path.join(dbgsym_root_dir, 'DEBIAN', 'control'), 0o644)


def _all_parent_directories_of(directories: Iterable[str]) -> Set[str]:
    result = {'.'}
    for path in directories:
        current = os.path.dirname(path)
        while current and current not in result:
            result.add(current)
            current = os.path.dirname(current)
    return result


def _auto_compute_multi_arch(binary_package: BinaryPackage,
                             package_root_dir: str,
                             fs_root: FSPath,
                             ) -> Optional[str]:
    resolved_arch = binary_package.resolved_architecture
    if resolved_arch == 'all':
        return None
    ctrl_dir = os.path.join(package_root_dir, 'DEBIAN')
    if any(script for script in ALL_CONTROL_SCRIPTS if os.path.isfile(os.path.join(ctrl_dir, script))):
        return None

    resolved_multiarch = binary_package.deb_multiarch
    assert resolved_arch != 'all'
    acceptable_no_descend_paths = {
        f'./usr/lib/{resolved_multiarch}',
        f'./usr/include/{resolved_multiarch}',
    }
    acceptable_files = {
        f'./usr/share/doc/{binary_package.name}/{basename}'
        for basename in (
            'copyright',
            'changelog.gz',
            'changelog.Debian.gz',
            f'changelog.Debian.{resolved_arch}.gz'
            'NEWS.Debian',
            'NEWS.Debian.gz',
            'README.Debian',
            'README.Debian.gz',
        )
    }
    acceptable_intermediate_dirs = _all_parent_directories_of(itertools.chain(acceptable_no_descend_paths,
                                                                              acceptable_files))

    for fs_path, children in fs_root.walk():
        path = fs_path.path
        if path in acceptable_no_descend_paths:
            children.clear()
            continue
        if path in acceptable_intermediate_dirs or path in acceptable_files:
            continue
        return None

    return 'same'


def _generate_control_files(binary_package: BinaryPackage,
                            package_state: PackageTransformationDefinition,
                            package_root_dir: str,
                            fs_root: FSPath,
                            substvars_file: str,
                            substvar_dependencies: Dict[str, Set[str]],
                            ) -> None:
    package = binary_package.name
    extra_common_params = []
    extra_params_specific = []
    with _ensure_base_substvars_defined(substvars_file) as substvars:
        if 'Installed-Size' not in substvars:
            # Pass it via cmd-line to make it more visible that we are providing the
            # value.  It also prevents the dbgsym package from picking up this value.
            ctrl_fs_root = build_fs_from_root_dir(os.path.join(package_root_dir, 'DEBIAN'))
            total_size = _compute_installed_size(fs_root) + _compute_installed_size(ctrl_fs_root)
            extra_params_specific.append(f'-VInstalled-Size={total_size}')
        for subsvar_name, dependencies in substvar_dependencies.items():
            for dependency in dependencies:
                substvars.add_dependency(subsvar_name, dependency)

    ma_value = binary_package.fields.get('Multi-Arch')
    if not binary_package.is_udeb and ma_value is None:
        ma_value = _auto_compute_multi_arch(binary_package, package_root_dir, fs_root)
        if ma_value is not None:
            print(f'The package "{binary_package.name}" looks like it should be "Multi-Arch: {ma_value}" based'
                  ' on the contents and there is no explicit "Multi-Arch" field. Setting the Multi-Arch field'
                  ' accordingly in the binary.  If this auto-correction is wrong, please  add "Multi-Arch: no" to the'
                  ' relevant part of "debian/control" to disable this feature.')
            extra_params_specific.append(f'-DMulti-Arch={ma_value}')
    elif ma_value == 'no':
        extra_params_specific.append(f'-UMulti-Arch')

    dbgsym_root_dir = dhe_dbgsym_root_dir(binary_package)
    dbgsym_ids = dhe_read_dbgsym_build_ids(binary_package)
    if package_state.binary_version is not None:
        extra_common_params.append(f'-v{package_state.binary_version}')

    if os.path.isdir(dbgsym_root_dir):
        _generate_dbgsym_control_file_if_relevant(binary_package,
                                                  dbgsym_root_dir,
                                                  substvars_file,
                                                  dbgsym_ids,
                                                  ma_value,
                                                  extra_common_params,
                                                  )
    elif dbgsym_ids:
        extra_common_params.append(f'-DBuild-Ids={dbgsym_ids}')

    dpkg_cmd = [
        'dpkg-gencontrol',
        f'-p{package}',
        '-ldebian/changelog',
        f'-T{substvars_file}',
        f'-P{package_root_dir}',
        *extra_common_params,
        *extra_params_specific,
    ]
    print(f'Executing for {package}: {dpkg_cmd}')
    subprocess.check_call(dpkg_cmd)
    os.chmod(os.path.join(package_root_dir, 'DEBIAN', 'control'), 0o644)
