import dataclasses
from enum import Enum
from typing import Callable, Optional


def _replace_dot_with_underscore(x: str) -> str:
    return x.replace('.', '_')


@dataclasses.dataclass(slots=True, frozen=True)
class PackagerProvidedFileClassSpec:

    lookup_basename: str
    install_into: str
    default_mode: int = 0o0644
    default_priority: Optional[int] = None
    filename_format: Optional[str] = None
    post_formatting_rewrite: Optional[Callable[[str], str]] = None

    @property
    def supports_priority(self) -> bool:
        return self.default_priority is not None


class PackagerProvidedFileClass(Enum):
    SYSTEMD_TMPFILES_FILE = PackagerProvidedFileClassSpec('tmpfiles', './usr/lib/tmpfiles.d',
                                                          filename_format='{name}.conf')
    PAM_CONFIG_FILE = PackagerProvidedFileClassSpec('pam', './usr/lib/pam.d')
    PPP_IP_UP_FILE = PackagerProvidedFileClassSpec('ppp.ip-up', './etc/ppp/ip-up.d', default_mode=0o0755)
    PPP_IP_DOWN_FILE = PackagerProvidedFileClassSpec('ppp.ip-down', './etc/ppp/ip-down.d', default_mode=0o0755)
    LOGROTATE_CONFIG_FILE = PackagerProvidedFileClassSpec('logrotate', './etc/logrotate.d')

    LOGCHECK_CRACKING_FILE = PackagerProvidedFileClassSpec(
        'logcheck.cracking',
        './etc/logcheck/cracking.d',
        post_formatting_rewrite=_replace_dot_with_underscore,
    )
    LOGCHECK_VIOLATIONS_FILE = PackagerProvidedFileClassSpec(
        'logcheck.violations',
        './etc/logcheck/violations.d',
        post_formatting_rewrite=_replace_dot_with_underscore,
    )
    LOGCHECK_VIOLATIONS_IGNORE_FILE = PackagerProvidedFileClassSpec(
        'logcheck.violations.ignore',
        './etc/logcheck/violations.ignore.d',
        post_formatting_rewrite=_replace_dot_with_underscore,
    )
    LOGCHECK_IGNORE_WORKSTATION_FILE = PackagerProvidedFileClassSpec(
        'logcheck.ignore.workstation',
        './etc/logcheck/ignore.d.workstation',
        post_formatting_rewrite=_replace_dot_with_underscore,
    )
    LOGCHECK_IGNORE_SERVER_FILE = PackagerProvidedFileClassSpec(
        'logcheck.ignore.server',
        './etc/logcheck/ignore.d.server',
        post_formatting_rewrite=_replace_dot_with_underscore
    )
    LOGCHECK_IGNORE_PARANOID_FILE = PackagerProvidedFileClassSpec(
        'logcheck.ignore.paranoid',
        './etc/logcheck/ignore.d.paranoid',
        post_formatting_rewrite=_replace_dot_with_underscore,
    )

    MIME_CONFIG_FILE = PackagerProvidedFileClassSpec('mime', './usr/lib/mime/packages/')
    SHARED_MIME_INFO_CONFIG_FILE = PackagerProvidedFileClassSpec('sharedmimeinfo',
                                                                 './usr/share/mime/packages',
                                                                 filename_format='{name}.xml',
                                                                 )

    IF_PRE_UP_CONFIG_FILE = PackagerProvidedFileClassSpec('if-pre-up', './etc/network/if-pre-up.d/',
                                                          default_mode=0o0755,
                                                          )
    IF_UP_CONFIG_FILE = PackagerProvidedFileClassSpec('if-up', './etc/network/if-up.d/', default_mode=0o0755)
    IF_DOWN_CONFIG_FILE = PackagerProvidedFileClassSpec('if-down', './etc/network/if-down.d/', default_mode=0o0755)
    IF_POST_DOWN_CONFIG_FILE = PackagerProvidedFileClassSpec('if-post-down', './etc/network/if-post-down.d/',
                                                             default_mode=0o0755,
                                                             )

    CRON_HOURLY_FILE = PackagerProvidedFileClassSpec('cron.hourly', './etc/cron.hourly', default_mode=0o0755)
    CRON_DAILY_FILE = PackagerProvidedFileClassSpec('cron.daily', './etc/cron.daily', default_mode=0o0755)
    CRON_WEEKLY_FILE = PackagerProvidedFileClassSpec('cron.weekly', './etc/cron.weekly', default_mode=0o0755)
    CRON_MONTHLY_FILE = PackagerProvidedFileClassSpec('cron.monthly', './etc/cron.monthly', default_mode=0o0755)
    CRON_YEARLY_FILE = PackagerProvidedFileClassSpec('cron.yearly', './etc/cron.yearly', default_mode=0o0755)
    # cron.d uses 0644 unlike the others
    CRON_D_CONFIG_FILE = PackagerProvidedFileClassSpec('cron.d', './etc/cron.d')

    INITRAMFS_HOOK_FILE = PackagerProvidedFileClassSpec('initramfs-hook', './usr/share/initramfs-tools/hooks',
                                                        default_mode=0o0755,
                                                        )

    MODPROBE_FILE = PackagerProvidedFileClassSpec('modprobe', './etc/modprobe.d', filename_format='{name}.conf')

    UDEV_FILE = PackagerProvidedFileClassSpec('udev', './lib/udev/rules.d', default_priority=60)

    GSETTINGS_OVERRIDE_FILE = PackagerProvidedFileClassSpec('gsettings-override', './usr/share/glib-2.0/schemas',
                                                            filename_format='{priority:02}_{name}.gschema.override',
                                                            default_priority=10)

    @property
    def default_lookup_basename(self) -> str:
        return self.spec.lookup_basename

    @property
    def install_into_directory(self) -> str:
        return self.spec.install_into

    @property
    def default_mode(self) -> int:
        return self.spec.default_mode

    def compute_dest_name(self, assigned_name: str, assigned_priority: Optional[int] = None) -> str:
        if assigned_priority is not None and not self.spec.supports_priority:
            raise ValueError(f'Cannot assign priority to packager provided files with stem'
                             f' "{self.spec.lookup_basename}" (e.g., "debian/foo.{self.spec.lookup_basename}"). They'
                             ' do not use priority at all.')

        filename_format = self.spec.filename_format
        if filename_format is None:
            filename_format = '{priority:02}-{name}' if self.spec.supports_priority else '{name}'
        if self.spec.supports_priority and assigned_priority is None:
            assigned_priority = self.spec.default_priority

        filename = filename_format.format(priority=assigned_priority,
                                          name=assigned_name,
                                          name_stem=self.spec.lookup_basename,
                                          )

        if self.spec.post_formatting_rewrite:
            filename = self.spec.post_formatting_rewrite(filename)
        return filename

    @property
    def spec(self) -> PackagerProvidedFileClassSpec:
        return self.value


PACKAGER_PROVIDED_FILE_DEFINITIONS = {
    x.default_lookup_basename: x for x in PackagerProvidedFileClass
}
